package consul

import (
	"fmt"
	"log"
	"strings"

	consulapi "github.com/hashicorp/consul/api"
)

// Config is configuration defined in the provider block
type Config struct {
	Datacenter    string `mapstructure:"datacenter"`
	Address       string `mapstructure:"address"`
	Scheme        string `mapstructure:"scheme"`
	HttpAuth      string `mapstructure:"http_auth"`
	Token         string `mapstructure:"token"`
	CAFile        string `mapstructure:"ca_file"`
	CAPem         string `mapstructure:"ca_pem"`
	CertFile      string `mapstructure:"cert_file"`
	CertPEM       string `mapstructure:"cert_pem"`
	KeyFile       string `mapstructure:"key_file"`
	KeyPEM        string `mapstructure:"key_pem"`
	CAPath        string `mapstructure:"ca_path"`
	InsecureHttps bool   `mapstructure:"insecure_https"`
	Namespace     string `mapstructure:"namespace"`
	client        *consulapi.Client
}

// Client returns a new client for accessing consul.
func (c *Config) Client() (*consulapi.Client, error) {
	config := consulapi.DefaultConfig()
	if c.Datacenter != "" {
		config.Datacenter = c.Datacenter
	}
	if c.Address != "" {
		config.Address = c.Address
	}
	if c.Scheme != "" {
		config.Scheme = c.Scheme
	}

	if c.CAFile != "" {
		config.TLSConfig.CAFile = c.CAFile
	}
	if c.CAPem != "" {
		config.TLSConfig.CAPem = []byte(c.CAPem)
	}
	if c.CertFile != "" {
		config.TLSConfig.CertFile = c.CertFile
	}
	if c.CertPEM != "" {
		config.TLSConfig.CertPEM = []byte(c.CertPEM)
	}
	if c.KeyFile != "" {
		config.TLSConfig.KeyFile = c.KeyFile
	}
	if c.KeyPEM != "" {
		config.TLSConfig.KeyPEM = []byte(c.KeyPEM)
	}
	if c.CAPath != "" {
		config.TLSConfig.CAPath = c.CAPath
	}
	if c.InsecureHttps {
		if config.Scheme != "https" {
			return nil, fmt.Errorf("insecure_https is meant to be used when scheme is https")
		}
		config.TLSConfig.InsecureSkipVerify = c.InsecureHttps
	}

	var err error
	config.HttpClient, err = consulapi.NewHttpClient(config.Transport, config.TLSConfig)
	if err != nil {
		return nil, fmt.Errorf("Failed to create http client: %s", err)
	}

	if c.HttpAuth != "" {
		var username, password string
		if strings.Contains(c.HttpAuth, ":") {
			split := strings.SplitN(c.HttpAuth, ":", 2)
			username = split[0]
			password = split[1]
		} else {
			username = c.HttpAuth
		}
		config.HttpAuth = &consulapi.HttpBasicAuth{Username: username, Password: password}
	}

	if c.Token != "" {
		config.Token = c.Token
	}

	client, err := consulapi.NewClient(config)

	log.Printf("[INFO] Consul Client configured with address: '%s', scheme: '%s', datacenter: '%s'"+
		", insecure_https: '%t'", config.Address, config.Scheme, config.Datacenter, config.TLSConfig.InsecureSkipVerify)
	if err != nil {
		return nil, err
	}
	return client, nil
}
